#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import sys

import pandas as pd


def entrez_db_list(input_list, output):
    """Function that stores all the taxa and accession is our database in a file"""

    db_list = []

    for fasta_file in input_list:
        path_list = fasta_file.split("/")
        taxon = path_list[-2]
        accession = path_list[-1].replace(".fasta.gz", "")
        db_list.append([taxon, accession])

    db_list_df = pd.DataFrame(db_list)
    db_list_df.to_csv(output, sep="\t", header=False, index=False)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    entrez_db_list(
        input_list=snakemake.input, output=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import sys
from collections import defaultdict

import pysam


def read_pair_generator(bam, region_string=None):
    """
    Generate read pairs in a BAM file or within a region string.
    Reads are added to read_dict until a pair is found.
    """
    read_dict = defaultdict(lambda: [None, None])
    for read in bam.fetch(region=region_string):

        if not read.is_proper_pair:
            continue

        qname = read.query_name
        if qname not in read_dict:
            if read.is_read1:
                read_dict[qname][0] = read
            else:
                read_dict[qname][1] = read
        else:
            if read.is_read1:
                yield read, read_dict[qname][1]
            else:
                yield read_dict[qname][0], read
            del read_dict[qname]


def mutation_type(alleles):
    """
    Is this mutation a transition (A <-> G and C <-> T) or a transversion (everything else).
    """
    alleles = set([a.upper() for a in alleles])
    assert len(alleles) == 2 and alleles.issubset({"A", "C", "G", "T", "N"})
    return "ts" if alleles in [{"A", "G"}, {"C", "T"}] else "tv"


def count_ts_tv_init(bam_file, output_file, taxon, pairs=False):
    bam = pysam.AlignmentFile(bam_file, "rb")

    with open(output_file, "w") as fout:

        for read in bam.fetch():
            ts, tv = 0, 0

            if not read.is_proper_pair:

                if pairs:
                    continue

                for base_call, base_ref in zip(read.seq, read.get_reference_sequence()):
                    if base_call.upper() == "N" or base_ref.upper() == "N":
                        continue

                    if base_call != base_ref:
                        if mutation_type([base_call, base_ref]) == "ts":
                            ts += 1
                        else:
                            tv += 1

                print(taxon, read.query_name, ts, tv, file=fout, sep=",")

        if pairs:

            for read1, read2 in read_pair_generator(bam):
                ts, tv = 0, 0

                for base_call1, base_ref1, base_call2, base_ref2 in zip(
                    read1.seq, read1.get_reference_sequence(), read2.seq, read2.get_reference_sequence(),
                ):

                    if (
                        base_call1.upper() == "N"
                        or base_ref1.upper() == "N"
                        or base_call2.upper() == "N"
                        or base_ref2.upper() == "N"
                    ):
                        continue

                    if base_call1 != base_ref1:
                        if mutation_type([base_call1, base_ref1]) == "ts":
                            ts += 1
                        else:
                            tv += 1

                    if base_call2 != base_ref2:
                        if mutation_type([base_call2, base_ref2]) == "ts":
                            ts += 1
                        else:
                            tv += 1

                print(taxon, read1.query_name, ts, tv, file=fout, sep=",")


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    count_ts_tv_init(
        bam_file=snakemake.input[0],
        output_file=snakemake.output[0],
        taxon=snakemake.wildcards.orgname,
        pairs=snakemake.params[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import argparse
import gzip
import sys

import pandas as pd
from Bio import SeqIO


def get_grey_matter_reads(input_fastq, matrix_file, output_fastq, matter):
    """Function to extract all the grey matter reads into a fastq file."""

    dirichlet_matrix = pd.read_csv(matrix_file, sep=",")

    ts_tv_group = dirichlet_matrix.groupby("Read_ID").sum().squeeze()

    if matter == "grey":
        grey_matter_df = ts_tv_group.where(ts_tv_group == 0)
        grey_matter_reads = grey_matter_df[grey_matter_df["Dirichlet_Assignment"].notna()].index.tolist()

        with gzip.open(input_fastq, "rt") as input_handle:
            with gzip.open(output_fastq, "wt") as output_handle:

                for record in SeqIO.parse(input_handle, "fastq"):

                    if record.id in grey_matter_reads:
                        SeqIO.write(record, output_handle, "fastq")

    # todo this is SLOW, it needs to be sped up -
    #  I think the main problem is writing to the file as the grey matter one is finishing much faster,
    #  and they both iterate through the same number of reads
    elif matter == "dark":
        aligned_read_names = ts_tv_group.index.tolist()
        with gzip.open(input_fastq, "rt") as input_handle:
            with gzip.open(output_fastq, "wt") as output_handle:

                dark_reads = []
                for record in SeqIO.parse(input_handle, "fastq"):

                    if record.id not in aligned_read_names:
                        dark_reads.append(record)
                        if len(dark_reads) >= 1000000:
                            SeqIO.write(dark_reads, output_handle, "fastq")
                            dark_reads = []


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Get the grey or dark matter reads..")

    parser.add_argument(
        "--input_fastq", help="Input fastq file to extract reads from.", metavar="",
    )
    parser.add_argument(
        "--matrix_file",
        help="Matrix file produced by haystac to that contains the posterior probabilities of each read belonging to "
        "a taxon in the db.",
        metavar="",
    )

    parser.add_argument(
        "--output_fastq", help="Output file path.", metavar="",
    )

    parser.add_argument(
        "--matter", help="Dark or grey matter reads.", metavar="",
    )

    args = parser.parse_args()

    # noinspection PyUnresolvedReferences
    get_grey_matter_reads(
        input_fastq=args.input_fastq, matrix_file=args.matrix_file, output_fastq=args.output_fastq, matter=args.matter
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import json
import os
import sys

import numpy as np
import pandas as pd

from haystac.workflow.scripts.utilities import PE


def calculate_likelihoods(ts_tv_file, readlen_file, taxa_file_paths, config, output_matrix, output_params):
    """
    Calculate the parameters for the analytical framework of the method, then the likelihoods for each read/taxon pair
    and perform the dirichlet distribution assignment.
    """
    assert os.stat(ts_tv_file).st_size, f"The ts_tv count file is empty {ts_tv_file}"
    assert os.stat(readlen_file).st_size, f"The read length is empty {readlen_file}"
    assert len(taxa_file_paths) > 0, f"The taxa list is empty {taxa_file_paths}"

    print("Reading the initial Ts/Tv matrix.", file=sys.stderr)
    init_ts_tv = pd.read_csv(ts_tv_file, names=["Taxon", "Read_ID", "Ts", "Tv"], sep=",")

    print("Calculating the sum of the transitions.", file=sys.stderr)
    ts_sum = int(init_ts_tv["Ts"].sum())

    print("Calculating the sum of the transversions.", file=sys.stderr)
    tv_sum = int(init_ts_tv["Tv"].sum())

    aligned_read_count = len(init_ts_tv["Read_ID"].unique())
    average_read_length = float(open(readlen_file, "r").read())

    max_mismatch = round(config["mismatch_probability"] * float(average_read_length))

    # calculate the parameters for the analytical framework
    if ts_sum == 0 and tv_sum != 0:
        sigma_t = 0
        sigma_v = config["mismatch_probability"]

        delta_t = 0
        delta_v = sigma_v / float(1 - sigma_v - sigma_t)

        ts_missing_val = 0
        tv_missing_val = max_mismatch

    elif ts_sum != 0 and tv_sum == 0:
        sigma_t = config["mismatch_probability"]
        sigma_v = 0

        delta_t = sigma_t / float(1 - sigma_v - sigma_t)
        delta_v = 0

        ts_missing_val = max_mismatch
        tv_missing_val = 0

    elif ts_sum == 0 and tv_sum == 0:
        sigma_t = config["mismatch_probability"] / (float(2))
        sigma_v = config["mismatch_probability"] / (float(2))

        delta_v = sigma_v / float(1 - sigma_v - sigma_t)
        delta_t = sigma_t / float(1 - sigma_v - sigma_t)

        ts_missing_val = round(max_mismatch / float(2))
        tv_missing_val = round(max_mismatch / float(2))

    else:
        ts_tv_ratio = ts_sum / float(tv_sum)
        fixed_mismatch_probability = config["mismatch_probability"]

        # solve linear system to find σv and σt
        coef = np.array([[1, 1], [1, -ts_tv_ratio]])
        const = np.array([fixed_mismatch_probability, 0])
        x = np.linalg.solve(coef, const)

        sigma_t = x[0]
        sigma_v = x[1]

        delta_v = sigma_v / float(1 - sigma_v - sigma_t)
        delta_t = sigma_t / float(1 - sigma_v - sigma_t)

        # solve linear system to find missing values for Ts and Tv
        coef_miss_tv = np.array([[1, 1], [1, -ts_tv_ratio]])
        const_miss_tv = np.array([max_mismatch, 0])
        y = np.linalg.solve(coef_miss_tv, const_miss_tv)

        ts_missing_val = round(y[0])
        tv_missing_val = round(y[1])

    # now calculate the likelihoods for each read/taxon pair
    data_ts_missing = pow(delta_t, ts_missing_val)
    data_tv_missing = pow(delta_v, tv_missing_val)

    # as we count max mismatches for every mate of a pair. So if a whole pair is missing we double the missing values
    if config["read_mode"] == PE:
        data_ts_missing = 2 * data_ts_missing
        data_tv_missing = 2 * data_tv_missing

    print("calculating the likelihood numerator", file=sys.stderr)
    init_ts_tv["ll_nom"] = init_ts_tv["Ts"].rpow(delta_t) * init_ts_tv["Tv"].rpow(delta_v)

    total_taxa_count = len(taxa_file_paths)

    print("calculating the proper likelihood", file=sys.stderr)
    init_ts_tv["Likelihood"] = np.nan

    init_ts_tv["Likelihood"] = init_ts_tv.groupby("Read_ID")["ll_nom"].transform(
        lambda nom: nom / (sum(nom) + ((total_taxa_count - nom.count()) * data_ts_missing * data_tv_missing))
    )

    print(init_ts_tv, file=sys.stderr)

    # every likelihood equal or above 0.95 gets assigned as 1, everything below as 0
    init_ts_tv["Dirichlet_Assignment"] = (
        init_ts_tv["Likelihood"]
        .where(init_ts_tv["Likelihood"] > config["min_prob"], 0.0)
        .where(init_ts_tv["Likelihood"] <= config["min_prob"], 1.0)
    )

    init_ts_tv.to_csv(output_matrix, sep=",", index=False)

    with open(output_params, "w") as fout:
        params = {
            "T": ts_sum,
            "V": tv_sum,
            "sigma_v": sigma_v,
            "sigma_t": sigma_t,
            "delta_v": delta_v,
            "delta_t": delta_t,
            "ts_missing_val": ts_missing_val,
            "tv_missing_val": tv_missing_val,
            "aligned_read_count": aligned_read_count,
        }

        json.dump(params, fout)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    calculate_likelihoods(
        ts_tv_file=snakemake.input[0],
        readlen_file=snakemake.input[1],
        taxa_file_paths=snakemake.input[2],
        config=snakemake.config,
        output_matrix=snakemake.output[0],
        output_params=snakemake.output[1],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os
import sys

import numpy as np
import pandas as pd
from scipy.stats import beta, hmean


def calculate_dirichlet_abundances(ts_tv_file, p_values_file, total_fastq_reads, sample_abundance):
    """
    Function that calculates the mean posterior abundances of species in metagenomic samples/libraries.
    """

    assert os.stat(ts_tv_file).st_size, f"The ts_tv count file is empty {ts_tv_file}"
    assert os.stat(p_values_file).st_size, f"The chi-square p values file is empty {p_values_file}"
    assert os.stat(total_fastq_reads).st_size, f"The total fastq reads file is empty {total_fastq_reads}"

    # I calculate the coverage of each taxon from reads in its bam/pileup file. Let's go there

    chi2_vector = (
        pd.read_csv(
            p_values_file, sep="\t", names=["species", "pvalue", "observed", "expected"], usecols=["species", "pvalue"]
        )
        .fillna(value=1)
        .groupby("species")
        .apply(hmean)
        .astype("float64")
        .rename("Taxon")
    )
    chi2_vector["Dark_Matter"] = np.nan
    chi2_vector["Grey_Matter"] = np.nan

    ts_tv_matrix = pd.read_csv(ts_tv_file, sep=",", usecols=["Taxon", "Read_ID", "Dirichlet_Assignment"])

    # Sum the Dirichlet Assignments per taxon and calculate the Dark Matter reads from the Dirichlet Assignment column
    ts_tv_group = ts_tv_matrix.groupby("Read_ID").sum().squeeze()
    grey_matter = ts_tv_group.where(ts_tv_group == 0).replace(0, 1).fillna(0)

    if len(ts_tv_matrix.Taxon.unique()) > 1:
        a = ts_tv_matrix.groupby("Taxon").sum().squeeze().astype(float)
    else:
        a = ts_tv_matrix.groupby("Taxon").sum().iloc[:, 0].astype(float)
    a.loc["Grey_Matter"] = grey_matter.sum()

    # Add the non aligned filtered reads count in the Dark Matter category
    total_fastq_reads = float(open(total_fastq_reads, "r").read())
    reads_in_bams = len(ts_tv_matrix["Read_ID"].unique())

    remaining_dark_matter = total_fastq_reads - reads_in_bams

    a.loc["Dark_Matter"] = remaining_dark_matter

    print(a, file=sys.stderr)

    # Perform Alberto's formulas
    b = a.sum()

    posterior_abundance_mean = a.add(1).divide(b + len(a)).sort_values(ascending=False)

    # Prepare the dataframe that is going to be outputted and calculate the rest of the output columns.
    posterior_abundance = posterior_abundance_mean.to_frame().reset_index()
    posterior_abundance.rename(columns={"Dirichlet_Assignment": "Mean_Posterior_Abundance"}, inplace=True)
    posterior_abundance["95_CI_lower"] = np.nan
    posterior_abundance["95_CI_upper"] = np.nan
    posterior_abundance["Minimum_Read_Num"] = np.nan
    posterior_abundance["Maximum_Read_Num"] = np.nan
    posterior_abundance["Dirichlet_Read_Num"] = np.nan
    posterior_abundance["Chi2_Contingency_Pvalue"] = np.nan

    # TODO is this necessary? if not, please remove and delete the log file redirection
    print(chi2_vector.index, file=sys.stderr)

    for idx, row in posterior_abundance.iterrows():
        ai = a.loc[posterior_abundance.iloc[idx, 0]]

        print(ai, file=sys.stderr)

        ci = beta.interval(0.95, ai + 1, b + len(a) - ai - 1)

        print(len(a), file=sys.stderr)

        posterior_abundance.iloc[idx, 2] = ci[0]
        posterior_abundance.iloc[idx, 3] = ci[1]
        posterior_abundance.iloc[idx, 4] = round(ci[0] * b)
        posterior_abundance.iloc[idx, 5] = round(ci[1] * b)
        posterior_abundance.iloc[idx, 6] = a.loc[posterior_abundance.iloc[idx, 0]]
        posterior_abundance.iloc[idx, 7] = chi2_vector.loc[posterior_abundance.iloc[idx, 0]]

    with open(sample_abundance, "w") as output_handle:
        posterior_abundance.to_csv(path_or_buf=output_handle, sep="\t", index=False, header=True)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    calculate_dirichlet_abundances(
        ts_tv_file=snakemake.input[0],
        p_values_file=snakemake.input[1],
        total_fastq_reads=snakemake.input[2],
        sample_abundance=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os
import sys

import pandas as pd
from scipy.stats import chi2_contingency


def coverage_chi2_contingency_test(cov_file, taxon_fasta_idx, taxon, outfile):
    """
    Function that calculates the pvalue of the coverage. Testing if there was clustering bias during the
    sequencing/the reads that contribute to the identification/abundance of a species come only
    from a specific genomic region
    """

    assert os.stat(cov_file).st_size, f"The file with the coverage stats {cov_file} is empty."

    cov_stats = pd.read_csv(cov_file, sep="\t", names=["observed", "expected"])

    expected_coverage = cov_stats['expected'].iloc[0]
    observed_coverage = cov_stats['observed'].iloc[0]

    contingency = [observed_coverage, expected_coverage]

    chi2, pvalue, dof, expected = chi2_contingency(contingency)

    with open(outfile, "w") as outhandle:
        print(taxon, pvalue, observed_coverage, expected_coverage, file=outhandle, sep="\t")


def genome_sizes(taxon_fasta_idx):
    faidx = pd.read_csv(
        taxon_fasta_idx,
        sep="\t",
        names=["Name", "Length", "Offset", "Linebases", "Linewidth"],
    )

    taxon_seq_len = faidx["Length"].sum()

    return taxon_seq_len


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    coverage_chi2_contingency_test(
        cov_file=snakemake.input[0],
        taxon_fasta_idx=snakemake.input[1],
        taxon=snakemake.wildcards.orgname,
        outfile=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import argparse
import hashlib
import os
import re
import sys

import pandas as pd
import yaml

REGEX_WHITELIST = r"[\w.-]+"
REGEX_BLACKLIST = r"[^\w.-]+"

PE = "PE"
COLLAPSED = "COLLAPSED"
SE = "SE"

WARNING = "\x1b[33m"
FAIL = "\x1b[31m"
END = "\033[0m"

is_tty = sys.stdout.isatty()


class ValidationError(Exception):
    pass


class ArgumentCustomFormatter(argparse.HelpFormatter):
    """
    Custom formatter for argparse
    """

    def _get_help_string(self, action):
        message = action.help
        if "%(default)" not in action.help:
            if action.default is not argparse.SUPPRESS and action.default is not None:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    message += " (default: %(default)s)"
        return message


class FileType(argparse.FileType):
    """
    Override argparse.FileType to return the filename, rather than an open file handle.
    """

    def __call__(self, string):
        return super().__call__(string).name


class WritablePathType(object):
    """
    Is this a writable path.
    """

    def __call__(self, value):
        from pathlib import Path

        try:
            path = Path(value).expanduser()
            path.mkdir(parents=True, exist_ok=True)
            return value
        except Exception:
            raise argparse.ArgumentTypeError(f"'{value}' is not a valid writable path")


class PositiveIntType(object):
    """
    Is this a positive integer
    """

    def __call__(self, value):
        try:
            if not int(value) > 0:
                raise ValueError()
        except ValueError:
            raise argparse.ArgumentTypeError(f"'{value}' is not a valid positive integer")

        return int(value)


class RangeType(object):
    """
    Is this a valid instance of `_type` and within the range [lower, upper]
    """

    def __init__(self, _type, lower, upper):
        self.type = _type
        self.lower = lower
        self.upper = upper

    def __call__(self, value):
        try:
            if not (self.lower <= self.type(value) <= self.upper):
                raise ValueError()
        except ValueError:
            raise argparse.ArgumentTypeError(
                f"'{value}' is not a valid {self.type.__name__} in the range ({self.lower}, {self.upper})"
            )

        return self.type(value)


class FloatRangeType(RangeType):
    """
    Is this a float() within the given range
    """

    def __init__(self, lower, upper):
        super().__init__(float, lower, upper)


class IntRangeType(RangeType):
    """
    Is this an int() within the given range
    """

    def __init__(self, lower, upper):
        super().__init__(int, lower, upper)


class BoolType(object):
    """
    Is this a valid boolean
    """

    def __call__(self, value):
        if isinstance(value, bool):
            return value
        if value.lower() in ("yes", "true", "t", "y", "1"):
            return True
        elif value.lower() in ("no", "false", "f", "n", "0"):
            return False
        else:
            raise argparse.ArgumentTypeError(f"'{value}' is not a valid boolean")


class JsonType(object):
    """
    Is this a valid JSON string
    """

    def __call__(self, value):
        import json

        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError as error:
            raise argparse.ArgumentTypeError(f"'{value}' is not a valid JSON string\n {error}")


class SpreadsheetFileType(object):
    """
    Is it a valid user input file
    """

    cols = None
    data = None

    def __call__(self, value):

        if not os.path.exists(value):
            raise argparse.ArgumentTypeError(f"'{value}' does not exit")

        if os.stat(value).st_size == 0:
            raise argparse.ArgumentTypeError(f"'{value}' is empty")

        try:
            self.data = pd.read_table(value, sep="\t", header=None, index_col=False,)
        except Exception:
            raise argparse.ArgumentTypeError(f"'{value}' unknown error parsing file")

        if len(self.data.columns) != len(self.cols):
            raise argparse.ArgumentTypeError(
                f"'{value}' must have {len(self.cols)} columns and be tab delimited (cols={len(self.data.columns)})"
            )

        # find the row number of any empty cells
        bad_rows = ", ".join([str(i + 1) for i in self.data.index[self.data.isnull().any(axis=1)].tolist()])

        if bad_rows:
            raise argparse.ArgumentTypeError(f"'{value}' contains missing data in line(s): {bad_rows}")

        return value


class AccessionFileType(SpreadsheetFileType):
    """
    Is this a valid accession input file.
    """

    cols = ["species", "accession"]

    def __call__(self, value):
        super().__call__(value)

        # check all accessions pass the regex pattern
        idx = self.cols.index("accession")
        species = self.cols.index("species")

        bad_list_acc = "\n".join(
            [
                f"line {i + 1}: '{acc}'"
                for i, acc in enumerate(self.data[idx].tolist())
                if re.match(REGEX_BLACKLIST, acc) is not None
            ]
        )

        bad_list_species = "\n".join(
            [
                f"line {i + 1}: '{tax}'"
                for i, tax in enumerate(self.data[species].tolist())
                if re.match(REGEX_BLACKLIST, tax) is not None
            ]
        )

        if bad_list_acc or bad_list_species:
            bad_list = bad_list_acc + "\n" + bad_list_species
            raise argparse.ArgumentTypeError(
                f"'{value}' these accession codes or taxon names contain invalid characters:\n{bad_list}"
            )

        return value


class SequenceFileType(AccessionFileType):
    """
    Is this a valid sequence input file.
    """

    cols = ["species", "accession", "path"]

    def __call__(self, value):
        super().__call__(value)

        # find any files that don't exist or are empty
        idx = self.cols.index("path")

        bad_files = "\n".join(
            [
                f"line {i + 1}: '{file}'"
                for i, file in enumerate(self.data[idx].tolist())
                if not os.path.exists(file) or os.stat(file).st_size == 0
            ]
        )

        if bad_files:
            raise argparse.ArgumentTypeError(f"'{value}' these sequence files do not exist or are empty:\n{bad_files}")

        return value


class SraAccessionType(object):
    """
    Is this a valid SRA accession
    """

    def __call__(self, value):
        # import these locally to avoid cyclic import issues
        from haystac.workflow.scripts.entrez_utils import entrez_esearch, entrez_efetch

        try:
            # query the SRA to see if this a valid accession
            _, _, id_list = entrez_esearch("sra", f"{value}[Accession]")
            etree = entrez_efetch("sra", id_list)
        except Exception:
            raise argparse.ArgumentTypeError(f"Invalid SRA accession '{value}'")

        try:
            # now get the library layout
            layout = etree.find(".//LIBRARY_LAYOUT/*").tag.lower()
        except Exception:
            raise argparse.ArgumentTypeError(f"Unable to resolve the library layout for SRA accession '{value}'")

        return value, layout


class NuccoreQueryType(object):
    """
    Is this a valid nuccore query
    """

    def __call__(self, value):
        # import these locally to avoid cyclic import issues
        from haystac.workflow.scripts.entrez_utils import entrez_esearch

        # check if the user has given us a file instead of a string
        if os.path.isfile(value):
            query = open(value).read().strip()
            if not query:
                raise argparse.ArgumentTypeError(f"The query file '{value}' is empty.")
        else:
            query = value

        try:
            # query nuccore to see if this a valid query
            _, _, id_list = entrez_esearch("nuccore", f"{query}")
        except Exception:
            raise argparse.ArgumentTypeError(f"Invalid NCBI query '{query}'")

        # if the query returns no result set raise error
        if len(id_list) == 0:
            raise argparse.ArgumentTypeError(f"No results in NCBI nucleotide for query '{query}'")

        return value


class CheckExistingConfig(object):
    """
    Checks the details of an existing yaml file against cli params or another yaml file
    """

    def __init__(self, filename, params):

        # check if second argument is a dict or a yaml file
        if isinstance(params, dict):
            params_config = params
        elif os.path.isfile(params):
            with open(params, "r") as fin_params:
                params_config = yaml.safe_load(fin_params)

        # check if a config already exists
        if not os.path.isfile(filename):
            pass

        else:
            # open the config file
            with open(filename, "r") as fin:
                existing_config = yaml.safe_load(fin)

                if not isinstance(params, dict):
                    important_args = ["cache"]
                else:
                    important_args = [
                        "cache",
                        "api_key",
                        "mismatch_probability",
                        "bowtie2_scaling",
                        "query",
                        "query_file",
                        "accessions_file",
                        "sequences_file",
                        "refseq_rep",
                        # "force_accessions",
                        # "exclude_accessions",
                        # "resolve_accessions",
                        "rank",
                        "mtDNA",
                        "seed",
                        "fastq",
                        "fastq_r1",
                        "fastq_r2",
                        "sra",
                        "collapse",
                        "trim_adapters",
                        "sample",
                        "min_prob",
                        "query_file_md5",
                        "accessions_md5",
                        "sequences_md5",
                    ]

                for arg in important_args:
                    # check if all the important params match
                    if arg in existing_config.keys() and arg in params_config.keys():
                        if existing_config[arg] != params_config[arg]:
                            print_error(
                                f"You are trying to set a value for parameter {arg} on top of an already existing one "
                                f"(old: {existing_config[arg]}, new: {params_config[arg]}). "
                                f"Please either revert to the original parameter you used or create a "
                                f"new output directory."
                            )


class FastqFile(object):
    """
    Is it a valid user input fastq file
    """

    def __call__(self, value):

        if not os.path.exists(value):
            raise argparse.ArgumentTypeError(f"'{value}' does not exit")

        if os.stat(value).st_size == 0:
            raise argparse.ArgumentTypeError(f"'{value}' is empty")

        if ".gz" not in value:
            with open(value, "r") as fin:
                first_line = fin.readline()
        else:
            import gzip

            with gzip.open(value, "rt") as fin:
                first_line = fin.readline()

        if first_line[0] != "@":
            raise argparse.ArgumentTypeError(f"'{value}' is not a valid fastq file.")

        return value


def get_total_paths(
    checkpoints, config,
):
    """
    Get all the individual fasta file paths for the taxa in our database.
    """

    sequences_df = pd.DataFrame()

    if config["query"]:
        pick_sequences = checkpoints.entrez_pick_sequences.get()
        sequences_df = pd.read_csv(pick_sequences.output[0], sep="\t")

        assert len(sequences_df) > 0, f"The entrez pick sequences file is empty {pick_sequences.output[0]}"

    if config["refseq_rep"]:
        refseq_rep_prok = checkpoints.entrez_refseq_accessions.get()
        refseq_genomes = pd.read_csv(refseq_rep_prok.output.refseq_genomes, sep="\t")
        genbank_genomes = pd.read_csv(refseq_rep_prok.output.genbank_genomes, sep="\t")
        assemblies = pd.read_csv(refseq_rep_prok.output.assemblies, sep="\t")
        refseq_plasmids = pd.read_csv(refseq_rep_prok.output.refseq_plasmids, sep="\t")
        genbank_plasmids = pd.read_csv(refseq_rep_prok.output.genbank_plasmids, sep="\t")

        if not config["force_accessions"]:
            invalid_assemblies = checkpoints.entrez_invalid_assemblies.get()
            invalid_assembly_sequences = pd.read_csv(invalid_assemblies.output[0], sep="\t")

            assemblies = assemblies[
                ~assemblies["AccessionVersion"].isin(invalid_assembly_sequences["AccessionVersion"])
            ]

        sources = [
            refseq_genomes,
            genbank_genomes,
            assemblies,
            refseq_plasmids,
            genbank_plasmids,
        ]

        if config["query"]:
            sources.append(sequences_df)

        sequences_df = pd.concat(sources)

    if config["sequences"] or config["accessions"]:
        check_unique_taxa_in_custom_inputs(config["accessions"], config["sequences"])

    if config["sequences"]:
        custom_fasta_paths = pd.read_csv(
            config["sequences"], sep="\t", header=None, names=["species", "AccessionVersion", "path"],
        )
        check_unique_taxa_in_user_file(custom_fasta_paths, config)

        custom_seqs = custom_fasta_paths[["species", "AccessionVersion"]].copy()
        custom_seqs["AccessionVersion"] = "custom_seq-" + custom_seqs["AccessionVersion"].astype(str)

        sequences_df = sequences_df.append(custom_seqs)

    if config["accessions"]:
        custom_accessions = pd.read_csv(
            config["accessions"], sep="\t", header=None, names=["species", "AccessionVersion"],
        )
        check_unique_taxa_in_user_file(custom_accessions, config)

        sequences_df = sequences_df.append(custom_accessions)

    if config["genera"]:
        sequences_df = sequences_df[sequences_df["species"].str.contains("|".join(config["genera"]))]

    if config["exclude_accessions"]:
        sequences_df = sequences_df[~sequences_df["AccessionVersion"].isin(config["exclude_accessions"])]

    inputs = []

    for key, seq in sequences_df.iterrows():
        orgname, accession = (
            normalise_name(seq["species"]),
            seq["AccessionVersion"],
        )
        inputs.append((orgname, accession))

    return inputs


def normalise_name(taxon):
    """remove unnecessary characters from a taxon name string."""

    return re.sub(REGEX_BLACKLIST, "_", taxon)


# return taxon.replace(" ", "_").replace("[", "").replace("]", "").replace("/", "_")


def check_unique_taxa_in_custom_inputs(accessions, sequences):
    """Checks that custom input files have only one entry per taxon"""

    if accessions != "" and sequences != "":
        custom_fasta_paths = pd.read_csv(sequences, sep="\t", header=None, names=["species", "accession", "path"])
        custom_accessions = pd.read_csv(accessions, sep="\t", header=None, names=["species", "accession"])

        taxon_acc = custom_accessions["species"].tolist()
        taxon_seq = custom_fasta_paths["species"].tolist()

        if bool(set(taxon_acc) & set(taxon_seq)):
            print_error(
                "You have provided the same taxon both in your custom sequences "
                "file and your custom accessions file. Please pick and keep ONLY "
                "one entry from both of these files. You can only have 1 sequence "
                "per chosen taxon in your database."
            )


def chunker(seq, size):
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def check_unique_taxa_in_user_file(df, config):
    """Checks that there are only unique inputs for taxa"""

    if df["species"].duplicated().any():
        dup_taxa = [i for i in df[df["species"].duplicated()]["species"].to_list()]
        message = f"{config['sequences']} contains multiple sequences for {', '.join(dup_taxa)}. "

        if not config["resolve_accessions"]:
            message += (
                "Either remove all duplicates, or set the `--resolve-accessions` flag to automatically choose one. "
                "It is the first accession that will be chosen."
            )
            print_error(message)
        else:
            df = df[~df["species"].duplicated(keep="first")]
            for idx, val in df[df["species"].duplicated()].iterrows():
                message += f"Accession {val['accession']} for {val['species']} was omitted."
                print_warning(message)
            return df


def md5(filename):
    hash_md5 = hashlib.md5()

    # open file and get the checksum
    with open(filename, "rb") as f:
        # read it in chunks in case the file is big
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


def print_error(message):
    """Function to print errors and exit"""
    message = f"haystac: error: {message}"
    print(f"{FAIL}{message}{END}" if is_tty else message, file=sys.stderr)
    exit(1)


def print_warning(message):
    """Function to print warnings"""
    message = f"WARNING: {message}"
    print(f"{WARNING}{message}{END}" if is_tty else message, file=sys.stderr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import gzip
import urllib.error
import urllib.request
from xml.etree import ElementTree

import requests
from Bio import bgzf

from haystac.workflow.scripts.entrez_utils import (
    entrez_range_accessions,
    entrez_request,
    entrez_assembly_ftp,
    entrez_find_replacement_accession,
    ENTREZ_MAX_UID,
    ENTREZ_MAX_ATTEMPTS,
)
from haystac.workflow.scripts.utilities import chunker, print_error


def download_entrez_ftp(ftp_url, output_file, attempt=1):
    """
    Read the FTP stream, unzip the contents and write them one line at a time to our bgzip file
    """
    try:
        # download the assembly file to a temp location
        temp_file, _ = urllib.request.urlretrieve(ftp_url)

        # stream the gzip file and recode as bgzip
        with gzip.open(temp_file) as fin, bgzf.open(output_file, "w") as fout:
            for line in fin:
                print(line.strip().decode("utf-8"), file=fout)

    except urllib.error.URLError as error:
        if attempt < ENTREZ_MAX_ATTEMPTS:
            # try downloading it again
            download_entrez_ftp(ftp_url, output_file, attempt + 1)
        else:
            raise error

    finally:
        # delete any temp files which are left behind by urlretrieve, and may fill up the `/tmp` folder
        urllib.request.urlcleanup()


def entrez_download_sequence(accession, output_file, force=False, mtdna=False):
    """
    Fetch the Entrez fasta record for a nuccore accession.
    """

    # query the assembly database to see if there is an FTP url we can use
    ftp_url = entrez_assembly_ftp(accession, force) if not mtdna else ""

    try:
        if ftp_url:
            download_entrez_ftp(ftp_url, output_file)
            return

    except urllib.error.URLError:
        pass

    try:
        # fetch the fasta record from nuccore
        r = entrez_request("efetch.fcgi", {"db": "nuccore", "id": accession, "rettype": "fasta", "retmode": "text"})

        # the fasta may be empty if this is a "master record" containing multiple other records (NZ_APLR00000000.1)
        if len(r.text.strip()) > 1:
            with bgzf.open(output_file, "w") as fout:
                print(r.text, file=fout)

            return

    except requests.exceptions.HTTPError:
        pass

    try:
        # get the full GenBank XML record
        r = entrez_request("efetch.fcgi", {"db": "nuccore", "id": accession, "rettype": "gb", "retmode": "xml"})

    except requests.exceptions.HTTPError:
        # check for a replacement accession (there may be a newer version if this a WGS project)
        updated_accession = entrez_find_replacement_accession(accession)

        # download the updated accession instead
        entrez_download_sequence(updated_accession, output_file, force)

        return

    # parse the XML result
    etree = ElementTree.XML(r.text)

    # get the first and last accession codes for this master record
    first = etree.find(".//GBAltSeqItem_first-accn")
    last = etree.find(".//GBAltSeqItem_last-accn")

    if first is None or last is None:
        print_error(f"Could not download the fasta file for {accession}")

    # get all the related accession codes
    accessions = entrez_range_accessions(accession, first.text, last.text)

    try:
        with bgzf.open(output_file, "w") as fout:
            # fetch all the accessions in batches
            for id_list in chunker(accessions, ENTREZ_MAX_UID):
                r = entrez_request(
                    "efetch.fcgi", {"db": "nuccore", "id": id_list, "rettype": "fasta", "retmode": "text"},
                )

                # write the fasta data to our bgzip file
                print(r.text, file=fout)

    except requests.exceptions.HTTPError:
        print_error(
            f"Could not download the accession range '{first.text}-{last.text}' for master record '{accession}'"
        )


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    entrez_download_sequence(
        accession=snakemake.wildcards.accession,
        output_file=snakemake.output[0],
        force=snakemake.config["force_accessions"],
        mtdna=snakemake.config["mtDNA"],
    )



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    required = [x for x in f.read().splitlines() if not x.startswith("#")]

setuptools.setup(
    name="haystac",
    version="0.2",
    author="Evangelos A. Dimopoulos, Evan K. Irving-Pease",
    author_email="antonisdim41@gmail.com",
    description="Species identification pipeline for both single species and metagenomic samples.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/antonisdim/haystac",
    package_dir={
        "haystac": "haystac",
        "haystac.workflow": "haystac/workflow",
        "haystac.workflow.scripts": "haystac/workflow/scripts",
    },
    packages=["haystac", "haystac.workflow", "haystac.workflow.scripts"],
    include_package_data=True,
    entry_points={"console_scripts": ["haystac = haystac.cli:Haystac"]},
    classifiers=["Programming Language :: Python :: 3", "License :: OSI Approved :: MIT License"],
    license="MIT",
    install_requires=required,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import sys

import pandas as pd
import pysam


def get_dirichlet_reads(input_bam, output_bam, matrix_file, taxon):
    """Function to get dirichlet assigned reads from a taxon's bam file"""

    dirichlet_matrix = pd.read_csv(matrix_file, sep=",")

    taxon_dirichlet_df = dirichlet_matrix[
        ((dirichlet_matrix["Taxon"] == taxon) & (dirichlet_matrix["Dirichlet_Assignment"] == 1))
    ]

    dirichlet_reads = taxon_dirichlet_df["Read_ID"].tolist()

    bam = pysam.AlignmentFile(input_bam, "rb")

    names = set(dirichlet_reads)

    with pysam.AlignmentFile(output_bam, "wb", header=bam.header) as fout:

        for read in bam.fetch():

            if read.query_name in names:
                fout.write(read)


if __name__ == "__main__":

    # noinspection PyUnresolvedReferences
    get_dirichlet_reads(
        input_bam=snakemake.input[0],
        output_bam=snakemake.output[0],
        matrix_file=snakemake.input[1],
        taxon=snakemake.wildcards.orgname,
    )
